/* 
	Copyright 2001 - Cycling '74
	Joshua Kit Clayton jkc@cycling74.com	
*/

#include "jit.common.h"

//for bumblebee

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <dc1394.h>


// change this to switch from RAW to JPEG
//#define VIDEO_MODE DC1394_VIDEO_MODE_FORMAT7_7
#define VIDEO_MODE  DC1394_VIDEO_MODE_FORMAT7_3
// number of frames to be recorded. The total number of files written to disk will be 24xNFRAMES.
#define NFRAMES 1
//open and close messages
#define CONNECT "open"
#define DISCONNECT "close"



typedef struct _jit_bumblebee 
{
	t_object				ob;
	long					mapcount;
	long					map[JIT_MATRIX_MAX_DIMCOUNT];
	t_symbol				*Uname;
	t_symbol				*matrix_name;
	
} t_jit_bumblebee;


//bumblebee
dc1394error_t err;
dc1394camera_t *camera;
dc1394video_frame_t *frame;
dc1394color_coding_t *color;
dc1394_t * d;
dc1394camera_list_t * list;
t_symbol *ps_done;



void *_jit_bumblebee_class;

t_jit_bumblebee *jit_bumblebee_new(void);
void jit_bumblebee_free(t_jit_bumblebee *x);
t_jit_err jit_bumblebee_matrix_calc(t_jit_bumblebee *x, void *inputs, void *outputs);
t_jit_err jit_bumblebee_init(void); 

t_jit_err jit_bumblebee_init(void) 
{
	
	long attrflags=0;
	void *mop, *o;
	
	_jit_bumblebee_class = jit_class_new("jit_bumblebee",(method)jit_bumblebee_new,(method)jit_bumblebee_free,
		sizeof(t_jit_bumblebee),0L); 

	//add mop
	mop = jit_object_new(_jit_sym_jit_mop,1,1); 
	jit_class_addadornment(_jit_bumblebee_class,mop);
	o=jit_object_method(mop,_jit_sym_getoutput,1);
	
	//add methods
	jit_class_addmethod(_jit_bumblebee_class, (method)jit_bumblebee_matrix_calc, 		"matrix_calc", 		A_CANT, 0L);
	
	//add attributes 
	attrflags = JIT_MATRIX_DATA_REFERENCE;
	
	//open camera code here
	if (!camera)
	{
		
		d = dc1394_new ();
		if (!d)
			error("could not open new camera instance");
		err=dc1394_camera_enumerate (d, &list);
		
		//DC1394_ERR_RTN(err,"Failed to enumerate cameras");
		if (err != 0) post("Failed to enumerate cameras: error code %d", err);
		
		if (list->num == 0) {
			error("No cameras found");
			return JIT_ERR_GENERIC;
		}
		camera = dc1394_camera_new (d, list->ids[0].guid);
		if (!camera) {
			error("Failed to initialize camera with guid %llx",list->ids[0].guid);
			return JIT_ERR_GENERIC;
		}
		
		dc1394_camera_free_list (list);
		
		post("Initializing \"%s %s\"...\n",camera->vendor,camera->model);
		
		// setup video mode, etc...
		err=dc1394_video_set_operation_mode(camera, DC1394_OPERATION_MODE_1394B);
		if (err != 0) post("Could not set B mode: err code %d", err);
		err=dc1394_video_set_iso_speed(camera, DC1394_ISO_SPEED_800);
		if (err != 0) post("Could not set 800Mbps speed: %d", err);
		err=dc1394_video_set_mode(camera, VIDEO_MODE);
		if (err != 0) post("Could not set video mode: %d", err);
		//tried to get color modes like this: err=dc1394_format7_get_color_coding(camera, VIDEO_MODE, color);
		//post("Available colormodes: %s\n", err);
		//	err=dc1394_format7_set_roi(camera, VIDEO_MODE, DC1394_COLOR_CODING_MONO8, 2000, 0,0, 1024, 4608);
		//	err=dc1394_format7_set_roi(camera, VIDEO_MODE, DC1394_COLOR_CODING_MONO8, 2000, 0,0, 1024, 4608);
		//err=dc1394_format7_set_roi(camera, VIDEO_MODE, DC1394_COLOR_CODING_RAW8, 2000, 0,0, 1024, 4608);
		err=dc1394_format7_set_roi(camera, VIDEO_MODE, DC1394_COLOR_CODING_RAW16, 2000, 0, 0, 640, 480);
		if (err != 0) post("Could not set ROI: %d", err);
		// setup capture
		err=dc1394_capture_setup(camera, 10, DC1394_CAPTURE_FLAGS_DEFAULT);
		if (err != 0) post("Could not setup capture: %d", err);
		err=dc1394_video_set_transmission(camera, DC1394_ON);
		if (err != 0) 
		{ 
			post("Could not start transmission: %d", err);
			
		} else if (err == 0) {
			post("\"%s %s\" successfully opened\n",camera->vendor,camera->model);
		}
	}

	jit_class_register(_jit_bumblebee_class);

	return JIT_ERR_NONE;
}

t_jit_err jit_bumblebee_matrix_calc(t_jit_bumblebee *x, void *inputs, void *outputs)
{
	t_jit_err err=JIT_ERR_NONE;
	long out_savelock;
	t_symbol stuff;
	long i;
	t_jit_matrix_info out_minfo,tmp_minfo;
	void *out_matrix,*tmp_matrix=NULL;
	t_matrix_conv_info mcinfo;
	
	out_matrix 	= jit_object_method(outputs,_jit_sym_getindex,0);
	
	err=dc1394_capture_dequeue(camera, DC1394_CAPTURE_POLICY_WAIT, &frame);
	//post("%X\n", &frame);
	if (err != 0) post("Could not dequeue a frame: error code %d\n", err);
	
	if (x&&out_matrix) {
		
		tmp_minfo.size = frame->allocated_image_bytes;
		tmp_minfo.dim[0] = 480;
		tmp_minfo.dim[1] = 640;
		tmp_minfo.dimstride[0] = 1;
		tmp_minfo.dimstride[1] = 1;
		tmp_minfo.dimcount = 2;
		tmp_minfo.planecount = 1;
		tmp_minfo.type = _jit_sym_char; ///< attribute type (char, long, float32, float64, symbol, atom, or obj)
		tmp_minfo.flags = JIT_MATRIX_DATA_REFERENCE|JIT_MATRIX_DATA_FLAGS_USE;
		mcinfo.flags = JIT_MATRIX_CONVERT_DSTDIM;
		mcinfo.srcdimstart[0] 	= 0;
		mcinfo.srcdimend[0] 	= tmp_minfo.dim[0]-1;
		mcinfo.srcdimstart[1] 	= 0;
		mcinfo.srcdimend[1] 	= tmp_minfo.dim[1]-1;

		out_savelock = (long) jit_object_method(out_matrix,_jit_sym_lock,1);
		char* output_data = NULL; jit_object_method(out_matrix,_jit_sym_getdata, &output_data);
		jit_object_method(out_matrix,_jit_sym_getinfo,&out_minfo);
		
		assert( out_minfo.dim[0] == 640 && out_minfo.dim[1] == 480 && out_minfo.dimcount == 2 && out_minfo.planecount == 4 );
		//assert( out_minfo.dim[0] == 640 && out_minfo.dim[1] == 480 && out_minfo.dimcount == 2 && out_minfo.planecount == 3 );
		
		// BAYER CORRECTION USING BASIC AVERAGING METHOD
		//err=dc1394_bayer_decoding_8bit(frame->image, output_data, out_minfo.dim[0], out_minfo.dim[1], DC1394_COLOR_FILTER_RGGB, DC1394_BAYER_METHOD_BILINEAR);
		//if (err != 0) post("Error from bayer = %d\n", err);
//		dc1394_bayer_decoding_8bit(const uint8_t *bayer, uint8_t *rgb,
//									uint32_t width, uint32_t height, dc1394color_filter_t tile,
//									dc1394bayer_method_t method);
			
		err = (t_jit_err) jit_object_method(out_matrix,_jit_sym_setinfo,&out_minfo);

		// release frame
		err=dc1394_capture_enqueue(camera, frame);
		if (err != 0) post("Could not enqueue a frame\n", err);
		
		if (err) goto out;
		
	} else {
		return JIT_ERR_INVALID_PTR;
	}
	
out:
	jit_object_method(out_matrix,_jit_sym_lock,out_savelock);
	if (tmp_matrix) jit_object_free(tmp_matrix);
	return err;
}

t_jit_bumblebee *jit_bumblebee_new(void)
{
	t_jit_bumblebee *x;
		
	if (x=(t_jit_bumblebee *)jit_object_alloc(_jit_bumblebee_class)) {
		//nada
	} else {
		x = NULL;
	}	
	return x;
}

void jit_bumblebee_free(t_jit_bumblebee *x)
{
	//nada
}
